#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

from .parser import (
    EhnLexer,
    EhnParser,
    EhnSyntaxError,
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`main-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

# pylint: disable=too-few-public-methods

from .base import (
    EhnParseEntityBase,
    EhnParseFeatureBase,
    EhnParseRestrictionBase,

    EhnParseFunctionHead,
    EhnParseStrHead,
    EhnParseValueBody
)

################################################################################################################################

class EhnParseNormalFeature(EhnParseFeatureBase, EhnParseStrHead, EhnParseValueBody):
    """E-HowNet Parsing: Normal Feature Node"""

    node_type = 'Feature'
    value_type = (EhnParseEntityBase, EhnParseRestrictionBase,)

    def __init__(self, head, value):
        EhnParseFeatureBase.__init__(self)
        EhnParseStrHead.__init__(self, head)
        EhnParseValueBody.__init__(self, value)

    def children(self):
        yield self.value

    def decode(self):
        return f'{self.head}={self.value.decode()}'

################################################################################################################################

class EhnParseFunctionFeature(EhnParseFeatureBase, EhnParseFunctionHead, EhnParseValueBody):
    """E-HowNet Parsing: Function Feature Node"""

    node_type = 'FunctionFeature'
    value_type = (EhnParseEntityBase, EhnParseRestrictionBase,)

    def __init__(self, function, value):
        EhnParseFeatureBase.__init__(self)
        EhnParseFunctionHead.__init__(self, function)
        EhnParseValueBody.__init__(self, value)

    def children(self):
        yield self.function
        yield self.value

    def decode(self):
        return f'{self.function.decode()}={self.value.decode()}'



#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author_name__ = 'Mu Yang'
__author_email__ = 'emfomy@gmail.com'
__copyright__ = '2018-2020 CKIP Lab'

__title__ = 'E-HowNet'
__version__ = '0.5.2'
__description__ = 'E-HowNet Utilities'
__license__ = "CC-BY-NC-SA 4.0"

__url__ = 'https://ehownet.readthedocs.io'
__download_url__ = __url__+'/tarball/'+__version__



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# pylint: disable=invalid-name, no-self-use

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

from itertools import (
    count as _count,
)

from .node import (
    EhnParseEntityBase as _EhnParseEntityBase,

    EhnParseCoindexEntity as _EhnParseCoindexEntity,
    EhnParseFunctionEntity as _EhnParseFunctionEntity,
    EhnParseTildeEntity as _EhnParseTildeEntity,
)

################################################################################################################################

LOGICAL_FUNCTIONS = [
    'union',
    'and',
    'or',
    'not',
    # 'Ques',
]

################################################################################################################################

def replace_tilde(root, coindex_prefix='x'):
    """Convert all tilde nodes to coindex nodes.

    Arguments
    ---------
        root : :class:`~ehn.parse.node.base.EhnParseNode`
            The output of :class:`~ehn.parse.parser.EhnParser`.

    """
    _TilderReplacer(root, coindex_prefix)
    return root

class _TilderReplacer:

    def __init__(self, root, coindex_prefix):
        coindexs = set(filter(None, map(self.get_coindex, root.descendant())))
        self.gen_coindex = (f'{coindex_prefix}{i}' for i in _count(1) if f'{coindex_prefix}{i}' not in coindexs)
        self.replace(root)

    def replace(self, node, *parents):
        if isinstance(node, _EhnParseTildeEntity):
            target = self.get_tilde_target(*parents)
            assert target, f'Can\'t find target for {node}'

            coindex = self.get_coindex(target)
            if not coindex:
                coindex = self.next_coindex()
                target.anchor.head = coindex
            node = _EhnParseCoindexEntity(coindex)

        # Recursive
        features = node.get_features()
        if features:
            node.features = [self.replace(feature, node, *parents) for feature in features]

        arguments = node.get_arguments()
        if arguments:
            node.arguments = [self.replace(argument, node, *parents) for argument in arguments]

        value = node.get_value()
        if value:
            node.value = self.replace(value, node, *parents)

        function = node.get_function()
        if function:
            node.function = self.replace(function, node, *parents)

        return node

    def get_tilde_target(self, *parents, count=0):
        if not parents:
            return None

        parent = parents[0]
        if isinstance(parent, _EhnParseEntityBase) and \
            not (isinstance(parent, _EhnParseFunctionEntity) and parent.head in LOGICAL_FUNCTIONS):
            count += 1
            if count > 1:
                return parent

        return self.get_tilde_target(parents[1:], count=count)

    def next_coindex(self):
        return next(self.gen_coindex)

    @staticmethod
    def get_coindex(node):
        anchor = node.get_anchor()
        if anchor and anchor.head:
            return anchor.head
        if isinstance(node, _EhnParseCoindexEntity):
            return node.head
        return None



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`main-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

# pylint: disable=too-few-public-methods

from .base import (
    EhnParseEntityBase,
    EhnParseFeatureBase,
    EhnParseFunctionBase,
    EhnParseRestrictionBase,
    EhnParseRootBase,

    EhnParseAnchorBody,
    EhnParseArgumentBody,
    EhnParseFeatureBody,
    EhnParseStrHead,
    EhnParseValueBody,
)

################################################################################################################################
# Root
#

class EhnParseRoot(EhnParseRootBase, EhnParseFeatureBody):
    """E-HowNet Parsing: Root Node"""

    node_type = 'Root'
    feature_type = EhnParseFeatureBase

    def __init__(self, *features):
        EhnParseRootBase.__init__(self)
        EhnParseFeatureBody.__init__(self, *features)

    @property
    def head(self):
        return self.features[0].head

    def children(self):
        yield from self.features

    def decode(self):
        return ','.join(feature.decode() for feature in self.features) if self.features else ''

################################################################################################################################
# Function
#

class EhnParseFunction(EhnParseFunctionBase, EhnParseArgumentBody, EhnParseStrHead):
    """E-HowNet Parsing: Function Node"""

    node_type = 'Function'
    argument_type = (EhnParseEntityBase, EhnParseRestrictionBase,)

    def __init__(self, head, *arguments):
        EhnParseFunctionBase.__init__(self)
        EhnParseArgumentBody.__init__(self, *arguments)
        EhnParseStrHead.__init__(self, head)

    def children(self):
        yield from self.arguments

    def decode(self):
        _arguments = ','.join(argument.decode() for argument in self.arguments)
        return f'{self.head}({_arguments})'

################################################################################################################################
# Restriction
#

class EhnParseRestriction(EhnParseRestrictionBase, EhnParseValueBody, EhnParseAnchorBody):
    """E-HowNet Parsing: Restriction Node"""

    node_type = 'Restriction'
    value_type = EhnParseEntityBase

    def __init__(self, value, anchor=None):
        EhnParseRestrictionBase.__init__(self)
        EhnParseValueBody.__init__(self, value)
        EhnParseAnchorBody.__init__(self, anchor)

    @property
    def head(self):
        return self.value.head

    def children(self):
        yield self.value

    def decode(self):
        return f'/{self.value.decode()}{self.anchor.decode()}'



#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

from setuptools import setup, find_namespace_packages
import ehn as about

################################################################################

def main():

    with open('README.rst') as fin:
        readme = fin.read()

    setup(
        name='ehownet',
        version=about.__version__,
        author=about.__author_name__,
        author_email=about.__author_email__,
        description=about.__description__,
        long_description=readme,
        long_description_content_type='text/x-rst',
        url=about.__url__,
        download_url=about.__download_url__,
        platforms=['linux_x86_64'],
        license=about.__license__,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3 :: Only',
            'License :: Free for non-commercial use',
            'Natural Language :: Chinese (Traditional)',
        ],
        packages=find_namespace_packages(include=['ehn', 'ehn.*',]),
        install_requires=[
            'ply>=3.11',
            'treelib>=1.5.5',
            'wcwidth>=0.1.7',
        ],
        entry_points={
            'console_scripts': [
                'ehn-parser = ehn._bin.parser:main',
            ],
        },
    )

################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`main-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

# pylint: disable=too-few-public-methods

from .base import (
    EhnParseEntityBase,
    EhnParseFeatureBase,

    EhnParseAnchorBody,
    EhnParseFeatureBody,
    EhnParseFunctionHead,
    EhnParseStrHead,
)

################################################################################################################################

class EhnParseNormalEntity(EhnParseEntityBase, EhnParseStrHead, EhnParseFeatureBody, EhnParseAnchorBody):
    """E-HowNet Parsing: Normal Entity Node"""

    node_type = 'Entity'
    feature_type = EhnParseFeatureBase

    def __init__(self, head, *features, anchor=None):
        EhnParseEntityBase.__init__(self)
        EhnParseStrHead.__init__(self, head)
        EhnParseFeatureBody.__init__(self, *features)
        EhnParseAnchorBody.__init__(self, anchor)

    def children(self):
        yield from self.features

    def decode(self):
        _features = ':' + ','.join(feature.decode() for feature in self.features) if self.features else ''
        return f'{{{self.head}{self.anchor.decode()}{_features}}}'

################################################################################################################################

class EhnParseFunctionEntity(EhnParseEntityBase, EhnParseFunctionHead, EhnParseFeatureBody, EhnParseAnchorBody):
    """E-HowNet Parsing: Function Entity Node"""

    node_type = 'FunctionEntity'
    feature_type = EhnParseFeatureBase

    def __init__(self, function, *features, anchor=None):
        EhnParseEntityBase.__init__(self)
        EhnParseFunctionHead.__init__(self, function)
        EhnParseFeatureBody.__init__(self, *features)
        EhnParseAnchorBody.__init__(self, anchor)

    def children(self):
        yield self.function
        yield from self.features

    def decode(self):
        _features = ':' + ','.join(feature.decode() for feature in self.features) if self.features else ''
        return f'{{{self.function.decode()}{self.anchor.decode()}{_features}}}'

################################################################################################################################

class EhnParseAnyEntity(EhnParseEntityBase):
    """E-HowNet Parsing: Any Entity Node"""

    node_type = 'AnyEntity'
    def __init__(self):
        EhnParseEntityBase.__init__(self)

    @property
    def head(self):
        return 'ANY'

    def children(self):
        return []

    @staticmethod
    def decode():
        return '{}'

################################################################################################################################

class EhnParseNameEntity(EhnParseEntityBase, EhnParseStrHead):
    """E-HowNet Parsing: Name Entity Node"""

    node_type = 'NameEntity'
    def __init__(self, head):
        EhnParseEntityBase.__init__(self)
        EhnParseStrHead.__init__(self, head)

    def children(self):
        return []

    def decode(self):
        return f'{{"{self.head}"}}'

################################################################################################################################

class EhnParseNumberEntity(EhnParseEntityBase, EhnParseStrHead):
    """E-HowNet Parsing: Number Entity Node"""

    node_type = 'NumberEntity'
    def __init__(self, head):
        EhnParseEntityBase.__init__(self)
        EhnParseStrHead.__init__(self, head)

    def children(self):
        return []

    def decode(self):
        return f'{{{self.head}}}'

################################################################################################################################

class EhnParseTildeEntity(EhnParseEntityBase):
    """E-HowNet Parsing: Tilde Entity Node

    .. deprecated:: 0.6

    """

    node_type = 'TildeEntity'
    def __init__(self):
        EhnParseEntityBase.__init__(self)

    @property
    def head(self):
        return '~'

    def children(self):
        return []

    @staticmethod
    def decode():
        return '{~}'

################################################################################################################################

class EhnParseCoindexEntity(EhnParseEntityBase, EhnParseStrHead):
    """E-HowNet Parsing: Coindex Entity Node"""

    node_type = 'CoindexEntity'

    def __init__(self, head):
        EhnParseEntityBase.__init__(self)
        EhnParseStrHead.__init__(self, head)

    def children(self):
        return []

    @property
    def _tree_label(self):
        return f'${self.head}'

    def decode(self):
        return f'{{{self.head}}}'



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`main-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

from .base import (
    EhnParseNode,
    EhnParseEntityBase,
    EhnParseFeatureBase,
    EhnParseFunctionBase,
    EhnParseRestrictionBase,
    EhnParseRootBase,
    EhnParseAnchor,
)

from .entity import (
    EhnParseNormalEntity,
    EhnParseFunctionEntity,
    EhnParseAnyEntity,
    EhnParseTildeEntity,
    EhnParseNameEntity,
    EhnParseNumberEntity,
    EhnParseCoindexEntity,
)

from .feature import (
    EhnParseNormalFeature,
    EhnParseFunctionFeature,
)

from .other import (
    EhnParseRoot,
    EhnParseFunction,
    EhnParseRestriction,
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`main-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

# pylint: disable=protected-access
# pylint: disable=too-few-public-methods

from abc import (
    ABCMeta as _ABCMeta,
    abstractmethod as _abstractmethod,
)

from collections import (
    defaultdict as _defaultdict,
)

from treelib import (
    Tree as _Tree,
)

################################################################################################################################
# Base
#

class EhnParseNode(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node"""

    def __init__(self):
        self.__tree = None

    @property
    @_abstractmethod
    def node_type(self):
        return NotImplemented

    @_abstractmethod
    def decode(self):
        return NotImplemented

    @property
    def _tree_label(self):
        _anchor = f' ${self.anchor.head}' if hasattr(self, 'anchor') and self.anchor.head else ''  # pylint: disable=no-member
        _text = f' {self._tree_label_text}' if hasattr(self, '_tree_label_text') else ''  # pylint: disable=no-member
        return f'[{self.node_type}{_anchor}]{_text}'

    def __str__(self):
        return f'<{self.__class__.__name__} {self.head}>'  # pylint: disable=no-member

    # def __str__(self):
    #     def write(line):
    #         nonlocal ret
    #         ret += line.decode() + '\n'
    #     ret = ''
    #     self.tree()._Tree__print_backend(data_property='_tree_label', func=write)
    #     return ret

    #################################################################################

    @_abstractmethod
    def children(self):
        return NotImplemented

    def descendant(self):
        yield self
        for child in self.children():
            yield from child.descendant()

    def tree(self):
        tree = EhnParseTree()
        self._create_tree(tree, None)
        return tree

    def _create_tree(self, tree, parent):
        idx = tree.create_node(parent=parent, data=self).identifier
        for child in self.children():
            child._create_tree(tree, idx)

    #################################################################################

    def get_features(self):
        return getattr(self, 'features', [])

    def get_arguments(self):
        return getattr(self, 'arguments', [])

    def get_value(self):
        return getattr(self, 'value', None)

    def get_function(self):
        return getattr(self, 'function', None)

    def get_anchor(self):
        return getattr(self, 'anchor', None)

################################################################################################################################
# Tree
#
class EhnParseTree(_Tree):

    def show(self, *args, data_property='_tree_label', **kwargs):  # pylint: disable=arguments-differ
        super().show(*args, data_property=data_property, **kwargs)


################################################################################################################################
# Base Types
#

class EhnParseEntityBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Entity Node"""

class EhnParseFeatureBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Feature Node"""

class EhnParseFunctionBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Function Node"""

class EhnParseRestrictionBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Function Node"""

class EhnParseRootBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Root Node"""

################################################################################################################################
# Anchor
#

class EhnParseAnchor:
    """E-HowNet Parsing: Node Anchor"""

    def __init__(self, head=None):
        self.head = head

    def __str__(self):
        return self.head if self.head else ''

    def __repr__(self):
        return str(self)

    def decode(self):
        return f'_{self.head}' if self.head else ''

################################################################################################################################
# Heads
#

class EhnParseStrHead(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with String Head"""

    def __init__(self, head):
        self.head = head

    @property
    def head(self):
        return self._head

    @head.setter
    def head(self, head):
        assert isinstance(head, str), f'‘{head}’ is not a str!'
        self._head = head  # pylint: disable=attribute-defined-outside-init

    @property
    def _tree_label_text(self):
        return self.head

################################################################

class EhnParseFunctionHead(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Function Head"""

    def __init__(self, function):
        self.function = function

    @property
    def head(self):
        return self.function.head

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, function):
        assert isinstance(function, EhnParseFunctionBase), f'‘{function}’ is not a Function!'
        self._function = function  # pylint: disable=attribute-defined-outside-init

################################################################################################################################
# Bodies
#

class EhnParseValueBody(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Value"""

    @property
    @_abstractmethod
    def value_type(self):
        return NotImplemented

    def __init__(self, value):
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        assert isinstance(value, self.value_type), f'‘{value}’ is not a {self.value_type}!'
        self._value = value  # pylint: disable=attribute-defined-outside-init

class EhnParseFeatureBody(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Feature"""

    @property
    @_abstractmethod
    def feature_type(self):
        return NotImplemented

    def __init__(self, *features):
        self._featuremap = _defaultdict(list)
        self.features = features

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, features):
        self._features = []  # pylint: disable=attribute-defined-outside-init
        for feature in features:
            self.add_feature(feature)

    def add_feature(self, feature):
        assert isinstance(feature, self.feature_type), f'‘{feature}’ is not a {self.feature_type}!'
        self._features.append(feature)
        self._featuremap[feature.head].append(feature)

    def __getitem__(self, key):
        return self._featuremap.get(key, [])

class EhnParseArgumentBody(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Argument"""

    @property
    @_abstractmethod
    def argument_type(self):
        return NotImplemented

    def __init__(self, *arguments):
        self.arguments = arguments

    @property
    def arguments(self):
        return self._arguments

    @arguments.setter
    def arguments(self, arguments):
        self._arguments = []  # pylint: disable=attribute-defined-outside-init
        for argument in arguments:
            self.add_argument(argument)

    def add_argument(self, argument):
        assert isinstance(argument, self.argument_type), f'‘{argument}’ is not a {self.argument_type}!'
        self._arguments.append(argument)

    def __getitem__(self, key):
        return self._arguments[key]

################################################################

class EhnParseAnchorBody(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Anchor"""

    def __init__(self, anchor=None):
        self.anchor = anchor or EhnParseAnchor()

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self, anchor):
        assert isinstance(anchor, EhnParseAnchor), f'‘{anchor}’ is not an Anchor!'
        self._anchor = anchor  # pylint: disable=attribute-defined-outside-init




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'COINDEX COLON COMMA EQUAL LBRACE LPAREN NUMBER QUOTE RBRACE RPAREN SLASH TEXT TILDE ULINEexpr : entity\n                | rootroot : feature\n                | root COMMA featureentityAny : LBRACE RBRACEentity : LBRACE NUMBER RBRACEentity : LBRACE COINDEX RBRACEentity : LBRACE QUOTE TEXT QUOTE RBRACEentity : LBRACE TILDE RBRACEentityOpen : LBRACE TEXTentityOpen : LBRACE functionentityAnchor : entityOpen anchorentityFeature : entityOpen   COLON feature\n                         | entityAnchor COLON featureentityFeature : entityFeature COMMA featureentity : entityOpen    RBRACE\n                  | entityAnchor  RBRACE\n                  | entityFeature RBRACEfeature : TEXT EQUAL entity\n                   | TEXT EQUAL entityAny\n                   | TEXT EQUAL restrictionfeature : function EQUAL entity\n                   | function EQUAL entityAny\n                   | function EQUAL restrictionfunction : TEXT LPAREN RPARENfunction : TEXT LPAREN restriction RPARENfunctionOpen : TEXT LPAREN entityfunctionArgument : functionOpen     COMMA entity\n                            | functionArgument COMMA entityfunction : functionOpen     RPAREN\n                    | functionArgument RPARENrestriction : SLASH entityrestriction : SLASH entity anchoranchor : ULINE COINDEX'
    
_lr_action_items = {'LBRACE':([0,20,21,30,32,34,44,],[4,43,4,43,4,4,4,]),'TEXT':([0,4,13,16,24,27,29,43,],[5,17,5,38,5,5,5,17,]),'$end':([1,2,3,9,22,26,28,35,36,37,39,40,41,42,49,52,53,54,58,59,61,62,],[0,-1,-2,-3,-16,-17,-18,-4,-6,-7,-9,-19,-20,-21,-34,-22,-23,-24,-5,-32,-8,-33,]),'COMMA':([3,8,9,11,12,22,26,28,35,36,37,39,40,41,42,47,48,49,50,51,52,53,54,55,56,58,59,61,62,],[13,29,-3,32,34,-16,-17,-18,-4,-6,-7,-9,-19,-20,-21,-27,-13,-34,-14,-15,-22,-23,-24,-28,-29,-5,-32,-8,-33,]),'NUMBER':([4,43,],[14,14,]),'COINDEX':([4,25,43,],[15,49,15,]),'QUOTE':([4,38,43,],[16,57,16,]),'TILDE':([4,43,],[18,18,]),'EQUAL':([5,10,31,33,45,60,],[20,30,-30,-31,-25,-26,]),'LPAREN':([5,17,],[21,21,]),'RBRACE':([6,7,8,14,15,17,18,19,22,23,26,28,31,33,36,37,39,40,41,42,43,45,48,49,50,51,52,53,54,57,58,59,60,61,62,],[22,26,28,36,37,-10,39,-11,-16,-12,-17,-18,-30,-31,-6,-7,-9,-19,-20,-21,58,-25,-13,-34,-14,-15,-22,-23,-24,61,-5,-32,-26,-8,-33,]),'COLON':([6,7,17,19,23,31,33,45,49,60,],[24,27,-10,-11,-12,-30,-31,-25,-34,-26,]),'ULINE':([6,17,19,22,26,28,31,33,36,37,39,45,59,60,61,],[25,-10,-11,-16,-17,-18,-30,-31,-6,-7,-9,-25,25,-26,-8,]),'RPAREN':([11,12,21,22,26,28,36,37,39,46,47,49,55,56,59,61,62,],[31,33,45,-16,-17,-18,-6,-7,-9,60,-27,-34,-28,-29,-32,-8,-33,]),'SLASH':([20,21,30,],[44,44,44,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expr':([0,],[1,]),'entity':([0,20,21,30,32,34,44,],[2,40,47,52,55,56,59,]),'root':([0,],[3,]),'entityOpen':([0,20,21,30,32,34,44,],[6,6,6,6,6,6,6,]),'entityAnchor':([0,20,21,30,32,34,44,],[7,7,7,7,7,7,7,]),'entityFeature':([0,20,21,30,32,34,44,],[8,8,8,8,8,8,8,]),'feature':([0,13,24,27,29,],[9,35,48,50,51,]),'function':([0,4,13,24,27,29,43,],[10,19,10,10,10,10,19,]),'functionOpen':([0,4,13,24,27,29,43,],[11,11,11,11,11,11,11,]),'functionArgument':([0,4,13,24,27,29,43,],[12,12,12,12,12,12,12,]),'anchor':([6,59,],[23,62,]),'entityAny':([20,30,],[41,53,]),'restriction':([20,21,30,],[42,46,54,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expr","S'",1,None,None,None),
  ('expr -> entity','expr',1,'p_expr','parser.py',150),
  ('expr -> root','expr',1,'p_expr','parser.py',151),
  ('root -> feature','root',1,'p_root','parser.py',156),
  ('root -> root COMMA feature','root',3,'p_root','parser.py',157),
  ('entityAny -> LBRACE RBRACE','entityAny',2,'p_entity_any','parser.py',166),
  ('entity -> LBRACE NUMBER RBRACE','entity',3,'p_entity_number','parser.py',170),
  ('entity -> LBRACE COINDEX RBRACE','entity',3,'p_entity_coindex','parser.py',174),
  ('entity -> LBRACE QUOTE TEXT QUOTE RBRACE','entity',5,'p_entity_name','parser.py',178),
  ('entity -> LBRACE TILDE RBRACE','entity',3,'p_entity_tilde','parser.py',182),
  ('entityOpen -> LBRACE TEXT','entityOpen',2,'p_entity_normal_open','parser.py',186),
  ('entityOpen -> LBRACE function','entityOpen',2,'p_entity_function_open','parser.py',190),
  ('entityAnchor -> entityOpen anchor','entityAnchor',2,'p_entity_anchor','parser.py',194),
  ('entityFeature -> entityOpen COLON feature','entityFeature',3,'p_entity_feature0','parser.py',199),
  ('entityFeature -> entityAnchor COLON feature','entityFeature',3,'p_entity_feature0','parser.py',200),
  ('entityFeature -> entityFeature COMMA feature','entityFeature',3,'p_entity_feature','parser.py',205),
  ('entity -> entityOpen RBRACE','entity',2,'p_entity_close','parser.py',210),
  ('entity -> entityAnchor RBRACE','entity',2,'p_entity_close','parser.py',211),
  ('entity -> entityFeature RBRACE','entity',2,'p_entity_close','parser.py',212),
  ('feature -> TEXT EQUAL entity','feature',3,'p_feature','parser.py',217),
  ('feature -> TEXT EQUAL entityAny','feature',3,'p_feature','parser.py',218),
  ('feature -> TEXT EQUAL restriction','feature',3,'p_feature','parser.py',219),
  ('feature -> function EQUAL entity','feature',3,'p_function_feature','parser.py',223),
  ('feature -> function EQUAL entityAny','feature',3,'p_function_feature','parser.py',224),
  ('feature -> function EQUAL restriction','feature',3,'p_function_feature','parser.py',225),
  ('function -> TEXT LPAREN RPAREN','function',3,'p_function_any','parser.py',230),
  ('function -> TEXT LPAREN restriction RPAREN','function',4,'p_function_restriction','parser.py',234),
  ('functionOpen -> TEXT LPAREN entity','functionOpen',3,'p_function_open','parser.py',238),
  ('functionArgument -> functionOpen COMMA entity','functionArgument',3,'p_function_argument','parser.py',242),
  ('functionArgument -> functionArgument COMMA entity','functionArgument',3,'p_function_argument','parser.py',243),
  ('function -> functionOpen RPAREN','function',2,'p_function_close','parser.py',248),
  ('function -> functionArgument RPAREN','function',2,'p_function_close','parser.py',249),
  ('restriction -> SLASH entity','restriction',2,'p_restriction','parser.py',254),
  ('restriction -> SLASH entity anchor','restriction',3,'p_restriction_anchor','parser.py',258),
  ('anchor -> ULINE COINDEX','anchor',2,'p_anchor','parser.py',263),
]



#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

import argparse

from ehn.parse import EhnParser, EhnSyntaxError

################################################################################################################################
# Main
#

def main(argv=None):

    argparser = argparse.ArgumentParser(description='E-HowNet Parser')

    argparser.add_argument('text', type=str, nargs='+', help='input texts.')
    argparser.add_argument('--debug', action='store_true', help='debug mode.')

    args = argparser.parse_args(argv)

    # lexer = EhnLexer()
    parser = EhnParser()

    for (i, text,) in enumerate(args.text):
        try:
            res = parser(text, debug=args.debug)
            print(f'#{i+1}')
            res.tree().show()
        except AssertionError as exc:
            print(exc)
            print(text)
        except EhnSyntaxError as exc:
            print(exc)
            print(text)
            print(exc.show_pos(text))



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# pylint: disable=invalid-name, no-self-use

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'

import re as _re
from wcwidth import wcswidth as _wcswidth
from ply.lex import lex as _lex
from ply.yacc import yacc as _yacc

from . import node as _node

################################################################################################################################
# Core
#

EHN_TOKENS_CHAR = {
    'QUOTE':  '"',
    'EQUAL':  '=',
    'COLON':  ':',
    'COMMA':  ',',
    'SLASH':  '/',
    'ULINE':  '_',
    'LPAREN': '(',
    'RPAREN': ')',
    'LBRACE': '{',
    'RBRACE': '}',
    'TILDE':  '~',
}

EHN_TOKENS = [
    'TEXT',
    'NUMBER',
    'COINDEX',
    *EHN_TOKENS_CHAR.keys()
]

class EhnSyntaxError(SyntaxError):
    """E-HowNet Syntax Error."""

    def __init__(self, *args, pos=None):
        super().__init__(*args)
        self.pos = pos

    def show_pos(self, text):
        """Show error position.

        Parameters
        ----------
        text : str
            original input text
        """
        return ' '*_wcswidth(text[:self.pos]) + '^'

################################################################################################################################
# Lexer
#

class _EhnLexer:

    def __init__(self, **kwargs):
        self._lexer = _lex(module=self, **kwargs)

    tokens = EHN_TOKENS

    # Define the lexer
    def t_ANY_error(self, t):
        raise EhnSyntaxError(f'Illegal character ‘{t.value[0]}’ at position {t.lexpos}.', pos=t.lexpos)
        # t.lexer.skip(1)

    # Skip all spaces
    # t_ignore = ' \t\n\r\f\v'

    # Default state tokens
    t_QUOTE = r'"'
    t_EQUAL = r'='
    t_COLON = r':'
    t_COMMA = r','
    t_SLASH = r'/'
    t_ULINE = r'_'
    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_LBRACE = r'{'
    t_RBRACE = r'}'
    t_TILDE = r'~'

    def t_TEXT(self, t):
        r'[A-Za-z0-9\x80-\U0010FFFF|+\-.?#]+'
        if _isnumber(t.value):
            t.type = 'NUMBER'
        if _is_coindex(t.value):
            t.type = 'COINDEX'
        return t

    # Invoke the lexer
    def __call__(self, data, *args, **kwargs):
        self._lexer.input(data)
        return iter(self._lexer)

class EhnLexer(_EhnLexer):
    """E-HowNet Lexer.

        .. method:: __call__(data)

            Run tokenization.
    """

################################################################################################################################
# Parser
#

class _EhnParser:

    def __init__(self, lexer=None, **kwargs):
        if lexer is not None:
            assert isinstance(lexer, EhnLexer), f'{lexer} is not an EhnLexer!'
            self.lexer = lexer
        else:
            self.lexer = EhnLexer()
        self._parser = _yacc(module=self, **kwargs)

    @property
    def _lexer(self):
        return self.lexer._lexer # pylint: disable=protected-access

    tokens = EHN_TOKENS

    # Define the parser
    def p_error(self, t):

        if t is None:
            msg = 'Unexpected ending.'
            pos = None
        else:
            msg = f'Unexpected symbol ‘{t.value}’ at position {t.lexpos}.'
            pos = t.lexpos

        syms = []
        for sym in self._parser.action[self._parser.state].keys():
            sym = EHN_TOKENS_CHAR.get(sym, sym)
            if sym == '$end':
                syms.append('‘ENDING’')
            else:
                syms.append(f'‘{sym}’')
        if len(syms) > 1:
            syms[-1] = 'or '+syms[-1]

        msg += f' Expecting a {", ".join(syms)}.'
        raise EhnSyntaxError(msg, pos=pos)

    # Object
    def p_expr(self, p):
        '''expr : entity
                | root'''
        p[0] = p[1]

    # Root
    def p_root(self, p):
        '''root : feature
                | root COMMA feature'''
        if len(p) == 2:
            p[0] = _node.EhnParseRoot(p[1])
        else:
            p[1].add_feature(p[3])
            p[0] = p[1]

    # Entity
    def p_entity_any(self, p):
        '''entityAny : LBRACE RBRACE'''
        p[0] = _node.EhnParseAnyEntity()

    def p_entity_number(self, p):
        '''entity : LBRACE NUMBER RBRACE'''
        p[0] = _node.EhnParseNumberEntity(p[2])

    def p_entity_coindex(self, p):
        '''entity : LBRACE COINDEX RBRACE'''
        p[0] = _node.EhnParseCoindexEntity(p[2])

    def p_entity_name(self, p):
        '''entity : LBRACE QUOTE TEXT QUOTE RBRACE'''
        p[0] = _node.EhnParseNameEntity(p[3])

    def p_entity_tilde(self, p):
        '''entity : LBRACE TILDE RBRACE'''
        p[0] = _node.EhnParseTildeEntity()

    def p_entity_normal_open(self, p):
        '''entityOpen : LBRACE TEXT'''
        p[0] = _node.EhnParseNormalEntity(p[2])

    def p_entity_function_open(self, p):
        '''entityOpen : LBRACE function'''
        p[0] = _node.EhnParseFunctionEntity(p[2])

    def p_entity_anchor(self, p):
        '''entityAnchor : entityOpen anchor'''
        p[1].anchor = p[2]
        p[0] = p[1]

    def p_entity_feature0(self, p):
        '''entityFeature : entityOpen   COLON feature
                         | entityAnchor COLON feature'''
        p[1].add_feature(p[3])
        p[0] = p[1]

    def p_entity_feature(self, p):
        '''entityFeature : entityFeature COMMA feature'''
        p[1].add_feature(p[3])
        p[0] = p[1]

    def p_entity_close(self, p):
        '''entity : entityOpen    RBRACE
                  | entityAnchor  RBRACE
                  | entityFeature RBRACE'''
        p[0] = p[1]

    # Feature
    def p_feature(self, p):
        '''feature : TEXT EQUAL entity
                   | TEXT EQUAL entityAny
                   | TEXT EQUAL restriction'''
        p[0] = _node.EhnParseNormalFeature(p[1], p[3])

    def p_function_feature(self, p):
        '''feature : function EQUAL entity
                   | function EQUAL entityAny
                   | function EQUAL restriction'''
        p[0] = _node.EhnParseFunctionFeature(p[1], p[3])

    # Function
    def p_function_any(self, p):
        '''function : TEXT LPAREN RPAREN'''
        p[0] = _node.EhnParseFunction(p[1])

    def p_function_restriction(self, p):
        '''function : TEXT LPAREN restriction RPAREN'''
        p[0] = _node.EhnParseFunction(p[1], p[3])

    def p_function_open(self, p):
        '''functionOpen : TEXT LPAREN entity'''
        p[0] = _node.EhnParseFunction(p[1], p[3])

    def p_function_argument(self, p):
        '''functionArgument : functionOpen     COMMA entity
                            | functionArgument COMMA entity'''
        p[1].add_argument(p[3])
        p[0] = p[1]

    def p_function_close(self, p):
        '''function : functionOpen     RPAREN
                    | functionArgument RPAREN'''
        p[0] = p[1]

    # Restriction
    def p_restriction(self, p):
        '''restriction : SLASH entity'''
        p[0] = _node.EhnParseRestriction(p[2])

    def p_restriction_anchor(self, p):
        '''restriction : SLASH entity anchor'''
        p[0] = _node.EhnParseRestriction(p[2], anchor=p[3])

    # Anchor
    def p_anchor(self, p):
        '''anchor : ULINE COINDEX'''
        p[0] = _node.EhnParseAnchor(p[2])

    # Invoke the parser
    def __call__(self, data, *args, debug=False, **kwargs):
        if debug:
            print(data)
            for tok in self.lexer(data):
                print(tok)
        ret = self._parser.parse(data, lexer=self._lexer, *args, debug=debug, **kwargs)
        return ret

class EhnParser(_EhnParser):
    """E-HowNet Parser.

        .. method:: __call__(data)

            Run parsing.
    """

################################################################################################################################
# Utility
#

def _isnumber(name):
    try:
        float(name)
        return True
    except ValueError:
        return False

def _is_coindex(name):
    return _is_coindex.pattern.match(name)
_is_coindex.pattern = _re.compile(r'x(\?|[0-9]*)')
